#!/usr/bin/env python3

"""
    =========================================================================================
 
        CS408 Environmental Monitoring Independent of Existing Infrastructure
        Copyright (C) 2021 Callum Inglis

        This program is free software; you can redistribute it and/or modify
        it under the terms of the GNU General Public License as published by
        the Free Software Foundation; either version 2 of the License, or
        (at your option) any later version.

        This program is distributed in the hope that it will be useful,
        but WITHOUT ANY WARRANTY; without even the implied warranty of
        MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
        GNU General Public License for more details.

        You should have received a copy of the GNU General Public License along
        with this program; if not, write to the Free Software Foundation, Inc.,
        51 Franklin Street, Fifth Floor, Boston, MA 02110-1301 USA.

        Contact: Callum.Inglis.2018(at)uni.strath.ac.uk

    =========================================================================================

    Libraries Used
        https://github.com/raspberrypi-tw/lora-sx1276

    =========================================================================================

    Usage: python Pi_Receiver.py -f 433 -b BW125 -s 7

    =========================================================================================
"""
import sys
from time import sleep, time
import datetime
import json
import requests  # pip3 install requests
from SX127x.LoRa import *
from SX127x.LoRaArgumentParser import LoRaArgumentParser
from SX127x.board_config import BOARD
import SX127x.packer as packer
import secrets
from multiprocessing import Process
import logging
import pika # pip3 install pika

sys.path.insert(0, '../')

# SYSTEM CONFIG
DEBUG = 1
API_URL = "https://" + secrets.API_HTTP_BASIC_AUTH + "emiei.4oh4.co/api"
MAX_TX_RESERVATION_TIME = 2  # Seconds
LOG_FILE = "Pi_Receiver.log"

# RabbitMQ Config
MQ_HOST = 'localhost'
MQ_USER = secrets.MQ_USER
MQ_PASS = secrets.MQ_PASS
MQ_EXCHANGE = 'EMIEI'
MQ_QUEUE_OUT = 'EMIEI-API-OUT'

# Configure Pi for LoRa
BOARD.setup()
parser = LoRaArgumentParser("Continuous LoRa receiver.")

class RxHello(object):
    """JSON Object received from sensor when they transmit a "TX Hello" request to send
    NOTE: Variable names must match JSON object from sensor!
    """

    def __init__(self, gatewayUID, uid, reservationTime, messageID):
        """Initialise object, we will pass JSON object from which these values are retrieved
        :param gatewayUID: UID Of this gateway
        :param uid: Sensor UID, of the form EMIEI-... where ... is a number
        :param reservationTime: Requested period of time to transmit message, may be overridden by ourselves
        :param messageID: ID of message coming from sensor
        """
        self.gatewayUid = gatewayUID
        self.uid = uid
        self.reservationTime = min(reservationTime, MAX_TX_RESERVATION_TIME)
        self.messageID = messageID
        self.okTransmit = False

    def set_ok(self, okTransmit):
        """Is sensor node OK to transmit?
        :param okTransmit: Boolean Value, True/False
        """
        self.okTransmit = okTransmit

    # Python Object to JSON Object
    def to_json(self):
        return json.dumps(self, default=lambda o: o.__dict__, sort_keys=True)


class SensorResponse(object):
    """JSON Object received from sensor when they transmit a "TX Payload"
    NOTE: Variable names must match JSON object from sensor!
    """

    def __init__(self, sensorMetadata, data):
        self.gatewayMetadata = GatewayMetadata()
        self.sensorMetadata = SensorMetadata(**sensorMetadata)
        self.sensorReading = SensorReading(**data)

    # Python Object to JSON Object
    def to_json(self):
        return json.dumps(self, default=lambda o: o.__dict__, sort_keys=True)

    """
    Publish Message to RMQ. 
    Consumer will pick up message and handle a reliable transmission to the API.
    """
    def send_to_api(self):
        try:
            connection = pika.BlockingConnection(pika.ConnectionParameters(host=MQ_HOST,credentials=pika.PlainCredentials(MQ_USER, MQ_PASS)))
            channel = connection.channel()
            channel.exchange_declare(exchange=MQ_EXCHANGE, exchange_type='direct', durable=True, auto_delete=False)
            channel.queue_declare(queue=MQ_QUEUE_OUT, durable=True, exclusive=False, auto_delete=False)
            channel.queue_bind(queue=MQ_QUEUE_OUT, exchange=MQ_EXCHANGE, routing_key=MQ_QUEUE_OUT)

            queueItem = {
                "timestamp" : datetime.datetime.today().strftime('%Y-%m-%d_%H-%M-%S'),
                "URL" : API_URL + "/sensor/reading",
                "DATA" : self.to_json(),
                "HEADERS" : {"Content-Type": "Application/json"}
            }

            channel.basic_publish(
                exchange=MQ_EXCHANGE, 
                routing_key=MQ_QUEUE_OUT, 
                body=json.dumps(queueItem)
            )

            return True

        except Exception as e:
            logging.error("[SensorResponse] Exception: %e" % e)
            return False
        

class GatewayMetadata(object):
    """JSON Object received from sensor when they transmit a "TX Payload". Component within SensorResponse(object)
    NOTE: Variable names must match JSON object from sensor!
    """
    def __init__(self):
        self.gatewayUID = get_gateway_serial()
        self.apiKey = secrets.GATEWAY_API_KEY


class SensorMetadata(object):
    """JSON Object received from sensor when they transmit a "TX Payload". Component within SensorResponse(object)
    NOTE: Variable names must match JSON object from sensor!
    """
    def __init__(self, uid, messageID, samplePeriod):
        self.uid = uid
        self.messageID = messageID
        self.samplePeriod = samplePeriod
        self.sampleTime = round(time())


class SensorReading(object):
    """JSON Object received from sensor when they transmit a "TX Payload". Component within SensorResponse(object)
    NOTE: Variable names must match JSON object from sensor!
    """
    def __init__(self, ppm, sht, co2):
        self.ppm = PPM(**ppm)
        self.sht = SHT(**sht)
        self.co2 = Co2(**co2)


class PPM(object):
    """JSON Object received from sensor when they transmit a "TX Payload".
    Component within SensorResponse(object).SensorReading
    NOTE: Variable names must match JSON object from sensor!
    """
    def __init__(self, p10, p25, p100):
        self.p10 = p10
        self.p25 = p25
        self.p100 = p100


class SHT(object):
    """JSON Object received from sensor when they transmit a "TX Payload".
    Component within SensorResponse(object).SensorReading
    NOTE: Variable names must match JSON object from sensor!
    """
    def __init__(self, temperature, humidity):
        self.temperature = temperature
        self.humidity = humidity


class Co2(object):
    """JSON Object received from sensor when they transmit a "TX Payload".
    Component within SensorResponse(object).SensorReading
    NOTE: Variable names must match JSON object from sensor!
    """
    def __init__(self, co2):
        self.co2 = co2

class FilterBySensorApiRequest(object):
    def __init__(self, sensorUID):
        self.apiKey = None
        self.sensorUID = sensorUID
        self.gatewayUID = get_gateway_serial()
        self.gatewayApiKey = secrets.GATEWAY_API_KEY

    # Python Object to JSON Object
    def to_json(self):
        return json.dumps(self, default=lambda o: o.__dict__, sort_keys=True)

def get_sensor_config_from_api(sensor_uid):
    headers = {'Content-Type': 'Application/json'}
    request_data = FilterBySensorApiRequest(sensor_uid).to_json()
    logging.info("[i] Fetching Sensor Config for %s. Request %s" % (sensor_uid, request_data))
    response = requests.post(API_URL + '/sensor/getConfig/', data=request_data, headers=headers)

    if response.status_code != 200:
        logging.info("[-] Failed to fetch config for Sensor %s, Status Code: %s" % (sensor_uid, response.status_code))
        return None

    logging.info("[+] Fetched Config for Sensor %s, Config: %s" % (sensor_uid, response.content))
    return json.loads(response.content)

class Reply:
    """JSON Object transmitted back to sensor when they transmit a "TX Payload".
    Include ack status, and any config updates, along with sensor ID, gateway ID & message ID
    NOTE: Variable names must match JSON object from sensor!
    """
    ackStatus = False

    def __init__(self, remoteSensorID, replyMsgID):
        self.gatewayUid = get_gateway_serial()
        self.uid = remoteSensorID
        self.replyMsgID = replyMsgID
        self.txAfterNReadings = None
        self.pollingFrequency = None

    def set_ack_status(self, ack_status):
        """Set Acknowledgement Status
        :param ack_status: True if gateway correctly processed value, False otherwise
        """
        self.ackStatus = ack_status

    def update_config_tx_after_n_readings(self, _n_readings):
        self.txAfterNReadings = _n_readings

    # Provide value in ms
    def update_config_polling_frequency(self, _polling_frequency):
        self.pollingFrequency = _polling_frequency

    def to_json(self):
        return json.dumps(self, default=lambda o: o.__dict__, sort_keys=True)


class LoRaReceiver(LoRa):
    """ See SX127x Library
    """
    def __init__(self, verbose=False):
        super(LoRaReceiver, self).__init__(verbose)
        self._id = get_gateway_serial()
        self.set_mode(MODE.SLEEP)
        self.set_dio_mapping([0] * 6)
        self.rx_hello_queue = RxHelloQueue()  # Track transmission requests

    def on_rx_done(self):
        self.clear_irq_flags(RxDone=1)
        payload = self.read_payload(nocheck=True)

        # Parse our data here!
        data = ''.join([chr(c) for c in payload])
        handle_incoming_lora_packet(self.rx_hello_queue, data)

        self.reset_ptr_rx()
        self.set_mode(MODE.RXCONT)

    def on_tx_done(self):
        logging.info("[LoRaReceiver] TxDone")
        # set RX
        self.set_dio_mapping([0, 0, 0, 0, 0, 0])  # RX
        sleep(1)
        self.reset_ptr_rx()
        self.set_mode(MODE.RXCONT)
        self.clear_irq_flags(RxDone=1)

    def on_cad_done(self):
        logging.info("on_CadDone")
        logging.info(self.get_irq_flags())

    def on_rx_timeout(self):
        logging.info("on_RxTimeout")
        logging.info(self.get_irq_flags())

    def on_valid_header(self):
        logging.info("on_ValidHeader")
        logging.info(self.get_irq_flags())

    def on_payload_crc_error(self):
        logging.info("on_PayloadCrcError")
        logging.info(self.get_irq_flags())

    def on_fhss_change_channel(self):
        logging.info("on_FhssChangeChannel")
        logging.info(self.get_irq_flags())

    def receive(self):
        self.reset_ptr_rx()
        self.set_mode(MODE.RXCONT)
        while True:
            if self.get_mode() == MODE.TX:
                rssi_value = self.get_rssi_value()
                status = self.get_modem_status()

                sys.stdout.flush()
                sys.stdout.write("\r%d %d %d" % (rssi_value, status['rx_ongoing'], status['modem_clear']))

    def transmit(self, _payload):
        global args
        self.tx_counter = 0
        self.write_payload(_payload)
        self.set_mode(MODE.TX)


def transmit_object_with_lora(obj):
    if DEBUG > 1:
        logging.info("[Transmit] %s" % obj.to_json())

    _length, _payload = packer.Pack_Str(obj.to_json())

    payload = [int(hex(c), 0) for c in _payload]

    sleep(0.1)  # Slight delay before transmitting, else too quick for sensor to start listening
    loraReceiver.transmit(payload)
    sleep(0.1)


class RxHelloQueue:
    """Keep track of what messages we have incoming and when their transmission window expires
    Used to accept / reject "TX Hello" requests
    """
    def __init__(self):
        self.reserved = False
        self.sensorID = None
        self.reservedUntil = None

    def has_reservation_expired(self):
        """Returns true when there is no longer any scheduled transmission due
        Returns false otherwise
        """
        logging.info("[RxHelloQueue] Time Now: %s, Reserved Until: %s" % (time(), self.reservedUntil))

        if self.reserved and self.reservedUntil < time():
            self.reserved = False
            return True

        return False

    def can_accept(self):
        self.has_reservation_expired()
        return not self.reserved

    def accept_new(self, sensor_id, reservation_duration):
        if not self.can_accept:
            return False

        self.reserved = True
        self.sensorID = sensor_id
        self.reservedUntil = time() + reservation_duration + 1

class MqConsumer:
    def __init__(self):
        self.connection = pika.BlockingConnection(pika.ConnectionParameters(host=MQ_HOST,credentials=pika.PlainCredentials(MQ_USER, MQ_PASS)))
        self.channel = self.connection.channel()
        self.channel.exchange_declare(exchange=MQ_EXCHANGE, exchange_type='direct', durable=True, auto_delete=False)
        self.channel.queue_declare(queue=MQ_QUEUE_OUT, durable=True, exclusive=False, auto_delete=False)
        self.channel.queue_bind(queue=MQ_QUEUE_OUT, exchange=MQ_EXCHANGE, routing_key=MQ_QUEUE_OUT)
        self.channel.basic_qos(prefetch_count=5)

        self.channel.basic_consume(MQ_QUEUE_OUT, self.callback, auto_ack=False)

    def run(self):
        logging.info('[MQ Consumer] [i] Waiting for messages')
        self.channel.start_consuming()

    # Consume message & try to send to API. Require 200 response
    def callback(self, ch, method, properties, body):
        content = json.loads(body.decode())
        data = json.loads(content['DATA'])
        logging.info(data['sensorMetadata'])
        logging.info(content)

        try:
            # If using self-signed cert turn off signature validation
            # response = requests.post(API_URL + '/sensor/reading', data=self.ToJson(), headers=headers, verify=False) #
            response = requests.post(content['URL'], data=content['DATA'], headers=content['HEADERS'])
        
        except Exception as e:
            logging.info("[MQ Consumer] [-] API Error (%s) During Transmission" % e)
            self.channel.basic_reject(delivery_tag=method.delivery_tag) # Re-Queue
            return

        # Already Reported
        if response.status_code == 208:
            logging.info("[MQ Consumer] [+] messageID %s (%s) from sensor %s already transmitted to API" % (data['sensorMetadata']['messageID'], content['timestamp'], data['sensorMetadata']['uid']))
            self.channel.basic_ack(delivery_tag=method.delivery_tag) # Ack from queue
            return

        # API Submissions Succesful
        if response.status_code >= 200 and response.status_code < 300:
            logging.info("[MQ Consumer] [+] Successfully transmitted messageID %s (%s) from sensor %s to API" % (data['sensorMetadata']['messageID'], content['timestamp'], data['sensorMetadata']['uid']))
            self.channel.basic_ack(delivery_tag=method.delivery_tag) # Ack from queue
            return

        # Unsucesful
        logging.info("[MQ Consumer] [-] API Error (%s) when transmitting messageID %s (%s) from sensor %s to API" % (response.status_code, data['sensorMetadata']['messageID'], content['timestamp'], data['sensorMetadata']['uid']))
        self.channel.basic_reject(delivery_tag=method.delivery_tag) # Re-Queue
        return

def get_gateway_serial():
    """Extract serial number from cpu info file
    Source: https://www.raspberrypi-spy.co.uk/2012/09/getting-your-raspberry-pi-serial-number-using-python/
    """
    cpu_serial = "0000000000000000"
    try:
        f = open('/proc/cpuinfo', 'r')
        for line in f:
            if line[0:6] == 'Serial':
                cpu_serial = line[10:26]
        f.close()

    except Exception:
        cpu_serial = "ERROR000000000"

    return cpu_serial


# Upon successfully receiving a message, send back an ack
def ack_message(sensor_response):
    data = Reply(sensor_response.sensorMetadata.uid, sensor_response.sensorMetadata.messageID)
    data.set_ack_status(True)

    # Retrieve up-to-date sensor config from API & Transmit back to sensor
    updated_sensor_config = get_sensor_config_from_api(sensor_response.sensorMetadata.uid)

    if updated_sensor_config is not None:
        data.update_config_polling_frequency(updated_sensor_config['pollingFrequency'])
        data.update_config_tx_after_n_readings(updated_sensor_config['txAfterNReadings'])

    transmit_object_with_lora(data)
    return


# When a sensor has data to transmit, it will send a TxHello. 
# Pick up that message here and decide if we can accept the message at this time.
def handle_rx_hello(rx_hello_queue, data):
    logging.info("[?] Check for RX Hello")
    if DEBUG > 1:
        logging.info("[i] Raw data: %s" % data)

    try:
        rx_hello_json = json.loads(data)
        rx_hello = RxHello(get_gateway_serial(), **rx_hello_json)

    except Exception as e:
        logging.error("\t[-] Not RX Hello")

        if DEBUG > 1:
            logging.error("\t[-] E: %s" % e)

        return False

    logging.info("    [+] RX Hello Confirmed!")

    # Check I'm not expecting any methods within rx_hello.reservationTime seconds
    if not rx_hello_queue.can_accept():
        rx_hello.set_ok(False)  # We're not actually going to send any reply, this improves overall system performance.
        logging.info("    [-] RX Hello \"No\". - Reserved by %s for %i seconds"
              % (rx_hello_queue.sensorID, rx_hello_queue.reservedUntil - time()))

    else:
        rx_hello_queue.accept_new(rx_hello.uid, rx_hello.reservationTime)

        # Send "OK" + UID
        rx_hello.set_ok(True)  # We are happy for this sensor to send its data
        transmit_object_with_lora(rx_hello)
        logging.info("    [+] RX Hello \"OK\" Reply Sent - %s is permitted to send for %d seconds"
              % (rx_hello.uid, rx_hello.reservationTime))

    return True


# handle SensorResponse data packet
def handle_sensor_response_packet(data):
    logging.info("[?] Check for Sensor Response Data")
    if DEBUG > 1:
        logging.info("[i] Raw data: %s" % data)

    try:
        parsed = json.loads(data)
        p = SensorResponse(**parsed)

    except Exception as e:
        logging.info("[-] Unable to Parse response, ignoring")

        if DEBUG > 1:
            logging.error("\t[-] E: %s" % e)

        return False

    logging.info("    [+] Sensor Response Data Confirmed!")

    if DEBUG > 1:
        logging.info("    [+] Received Payload: %s" % p.to_json())

    # Send Data To API
    logging.info("    [+] Sending Data to API")
    if p.send_to_api():
        logging.info("    [+] Data Sent to API")

        logging.info("    [+] Sending Sensor Response Ack")
        ack_message(p) # Transmit ACK bak to sensor
        logging.info("    [+] Sensor Response Ack Sent")

    else:
        logging.error("    [-] Failed to send data to API") # No not Ack! Sensor will re-transmit

    sleep(0.1)
    return True


# Parse LoRa response, validate, save / transmit
def handle_incoming_lora_packet(rx_hello_queue, data):
    logging.info("[i] Packet Incoming")

    """  Handle TX Hello Transmission
          Rx - "I've got data to send!" + UID (From Sensor)
          
          If not expecting any messages, permit payload send, else send tx auth denied 
          Tx - "OK" + UID (Back to Sensor) + TX_Auth_ID
     [or] TX - "NO" + UID (Back to Sensor)
    """
    if handle_rx_hello(rx_hello_queue, data):  # Handled all OK
        return

    """ Handle Payload Transmission
          Rx - Packet + UID + Tx_Auth_ID (From Sensor)
          
          Handle message, ack / nak appropriately
          Tx - Ack + UID + Next_Send_Interval
     [or] Tx - Nak + UID
    """
    if handle_sensor_response_packet(data):
        return

if __name__ == '__main__':
    logging.basicConfig(level=logging.INFO, filename=LOG_FILE, filemode="a+", format="%(asctime)-15s %(levelname)-8s %(message)s")  
    logging.getLogger("pika").setLevel(logging.WARNING) # Reduce logging level

    # Setup Receiver
    loraReceiver = LoRaReceiver(verbose=False)
    args = parser.parse_args(loraReceiver)

    loraReceiver.set_mode(MODE.STDBY)
    loraReceiver.set_pa_config(pa_select=PA_SELECT.PA_BOOST)
    loraReceiver.set_pa_config(max_power=300, output_power=500)  # Default max_power = 10.8dB, output_power  = 4.2dB

    logging.info("[+] Receiver & API Gateway")
    logging.info("[i] Power Levels: %s" % (loraReceiver.get_pa_config(convert_dBm=True)))

    loraReceiver.set_agc_auto_on(1)
    assert (loraReceiver.get_agc_auto_on() == 1)

    # Start Consumer
    subscriber_list = []
    subscriber_list.append(MqConsumer())

    process_list = []
    for sub in subscriber_list:
        process = Process(target=sub.run)
        process.start()
        process_list.append(process)

    try:
        loraReceiver.receive()

    except KeyboardInterrupt:
        sys.stdout.flush()
        print("")
        sys.stderr.write("KeyboardInterrupt\n")

    finally:
        sys.stdout.flush()
        print("")
        loraReceiver.set_mode(MODE.SLEEP)
        BOARD.teardown()

    # wait for finish
    for process in process_list:
        process.join()
